// ir2ir.cpp : ircodes converter

#include <stdio.h>
#include <math.h>
#include <string.h>

#define SAPIEN2_CMD_COUNT	73
#define TYPE_V1HEX	1
#define TYPE_V2HEX	2
#define TYPE_V1CMD	3
#define TYPE_V2CMD	4
#define TYPE_UIRT	5
#define TYPE_PRONTO	6

static double unit;
static char *cmd[128] = { "turn_right", "right_arm_up", "right_arm_out", "tilt_body_right",
	                "right_arm_down", "right_arm_in", "walk_forward", "walk_backward",
					"turn_left", "left_arm_up", "left_arm_out", "tilt_body_left",
					"left_arm_down", "left_arm_in", "stop", "-", "master_program",
					"play", "right_program", "left_program", "sonic_program",
					"-", "-", "-", "quiet_execute", "-", "quiet_execute_with_subroutines", "-",
					"-", "-", "-", "-",
					"right_turn_step", "right_hand_thump", "right_hand_throw",
					"sleep", "right_hand_pickup", "lean_backward", "forward_step",
					"backward_step", "left_turn_step", "left_hand_thump",
					"left_hand_throw", "listen", "left_hand_pickup", "lean_forward" ,
					"reset", "-", "master_execute", "wakeup", "right_execute", "left_execute",
					"sonic_execute", "-", "-", "-", "-", "-", "-", "-", "-", "-", "-", "-",
					"right_hand_strike_3", "right_hand_sweep",
					"burp", "right_hand_strike", "high_5", "right_hand_strike_1",
					"bulldozer", "oops", "left_hand_strike_3", "left_hand_sweep",
					"whistle", "left_hand_strike_2", "talkback", "left_hand_strike_1",
					"roar", "-", "all_demo", "power_off", "demo_1", "demo_2", "dance", "-", "karate_chop",
					"-","-","-","-","-","-","-","-","-","-","-","-","-","-","-","-","-","-","-","-",
					"-","-","-","-","-","-","-","-","-","-","-",
					"feet_shuffle", "-","-","-","-","nothing", "raise_arm_throw","-","-","-"  };

static char *cmd2[SAPIEN2_CMD_COUNT] = { "_walking",
                        "stop",
						"sonic_sensors_on_off",
						"vision_systems_on_off",
						"positional_program_entry",
						"positional_program_play",
						"gait_change_quit_reset",
						"free_roam",
						"dance_demo",
						"_head_and_upper_body",
						"reset",
						"right_arm_throw",
						"right_arm_low_pickup",
						"right_arm_high_pickup",
						"right_arm_grab",
						"right_arm_give",
						"right_arm_roll",
						"movement_demo",
						"_right_arm",
						"sleep_wake_up",
						"left_arm_throw",
						"left_arm_low_pickup",
						"left_arm_high_pickup",
						"left_arm_grab",
						"left_arm_give",
						"left_arm_roll",
						"lie_down_sit_up_lie_down_stand_up",
						"_left_arm",
						"get_up",
						"right_kick",
						"right_push",
                        "right_chop",
						"left_chop",
						"left_push",
						"left_kick",
						"oops",
						"_hip_and_waist_tilt",
						"clear_program",
						"sound_program",
						"vision_program",
						"main_program",
						"perform_program",
						"guard_mode",
						"clear_entry",
						"high_5",
						"_both_arms",
						"bulldozer_forward",
						"laugh",
						"insult",
						"right_arm_drop",
						"left_arm_drop",
						"plan",
						"spare_change",
						"hey_baby",
						"_head_only",
						"bulldozer_backward",
						"roar",
						"diode",
						"fetch",
						"danger",
						"calm_down",
						"hug",
						"burp",
						"_head_and_upper_body",
						"power_down",
						"robosapien_v2_interaction",
						"roboraptor_interaction",
						"robopet_interaction",
						"manual_color_mode_daylight",
						"manual_color_mode_indoor_yellow",
						"manual_color_mode_indoor_white",
						"dont_press",
						"_upper_body_only" };

static int codes2[SAPIEN2_CMD_COUNT] = { 0x300, 0x3aa, 0x383, 0x380, 0x381, 0x352, 0x390, 0x382, 0x350,
                          0x320, 0x353, 0x355, 0x356, 0x357, 0x358, 0x359, 0x35a,
						  0x354, 0x328, 0x386, 0x35c, 0x35d, 0x35e, 0x35f, 0x360,
						  0x361, 0x351, 0x330, 0x364, 0x362, 0x365, 0x366, 0x367,
						  0x368, 0x363, 0x35b, 0x340, 0x3f4, 0x3f1, 0x3f0, 0x3f5,
						  0x3f6, 0x3f2, 0x3f3, 0x369, 0x338, 0x36a, 0x36c, 0x36d,
						  0x36e, 0x36f, 0x370, 0x371, 0x36b, 0x348, 0x372, 0x374,
						  0x375, 0x376, 0x377, 0x378, 0x379, 0x373, 0x310, 0x384,
                          0x385, 0x387, 0x388, 0x389, 0x38a, 0x38b, 0x37a, 0x318 };



int sapien_decode(char *sapien_cmd, int version)
{
  int i;
  if (version == 1)
  {
    for (i = 0; i < 128; i++)
	  if (strcmp(cmd[i], sapien_cmd) == 0)
		  return 128 + i;
  }
  else // sapien v2
  {
    int dir = 0;
	char *s;
    if (sapien_cmd[0] == 'D')
	{
		s = sapien_cmd + 2;
		dir = sapien_cmd[1] - '0';
	}
	else s = sapien_cmd;
	for (i = 0; i < SAPIEN2_CMD_COUNT; i++)
	  if (strcmp(cmd2[i], s) == 0)
	    return codes2[i] + dir;
  }
  printf("ir2ir: unrecognized command: %s\n", sapien_cmd);
  return 0;
}

int hexdigit(char *d)
{
	if (*d > 'a') return 0xa + (*d - 'a');
	else if (*d > 'A') return 0xa + (*d - 'A');
	else return *d - '0';
}

char *sapienhex2uirt(int b, int datalength)
{
  static char rv[6*32+10];
  if ((datalength > 32) || (datalength < 1)) return 0;
  unsigned long x = ((unsigned long)1) << (unsigned long)(datalength - 1); 
  
  if (b <= 0) return "unrecognized";

  sprintf(rv, "R08008104");
  char *p = rv + 9;

  while (datalength--)
  {
    if (b & x) 
	{
	  sprintf(p, "808221");
	  p += 6;
	}
	else
	{
	  sprintf(p, "2121");
	  p += 4;
	}
	x >>= 1;
  }
  return rv;
}

char *hex2pronto(int code, int version)
{
	static char rv[160];
    char *p;

	if (code <= 0) return "unrecognized";
	//0x6A = 106 = (int) 4145146 / 39200   where 4145146Hz is pronto internal clock 
	//                                     freq. and 39.2kHz is Sapien IR carrier frequency
	sprintf(rv, "0000 006A 0000 ");
	p = rv + 15;
	if (version == 1) sprintf(p, "0009 ");   //version 1: 9 bursts
	else sprintf(p, "000D "); //version 2: 13 bursts
	p += 5;

	//signal header
	sprintf(p, "0105 ");
	p += 5;
	//process the data bits
	unsigned long x = (version == 1) ? 128 : 2048;
	int iter = (version == 1) ? 8 : 12;

    while (iter--)
    {
      if (code & x) 
	  {
	    sprintf(p, "0082 0021 ");
	    p += 10;
	  }
	  else
	  {
	    sprintf(p, "0021 0021 ");
	    p += 10;
	  }
	  x >>= 1;
    }
    return rv;
}

char *uirt2pronto(char *uirt)
{
	static char rv[160];
    char *p;

	//0x6A = 106 = (int) 4145146 / 39200   where 4145146Hz is pronto internal clock 
	//                                     freq. and 39.2kHz is Sapien IR carrier frequency
	sprintf(rv, "0000 006A 0000 0000 ");
	p = rv + 20;

	//process the uirt data
	if (uirt[0] != 'R') return "only R-type UIRT packets are recognized";

	char *q = uirt + 1;
	
	int n = 0;
    while ((*q != 0) && (*q != '\n'))
    {
	  int val = (hexdigit(q) << 4) | hexdigit(q + 1);
	  q += 2;
	  if (val > 0x7f) 
	  {
		  val = ((val & 0x7f) << 8) | 
		        (hexdigit(q) << 4) | hexdigit(q + 1);
		  q += 2;
	  }

	  sprintf(p, "%x%x%x%x ", val >> 12, 0xF & (val >> 8), 0xF & (val >> 4), val & 0xF);
	  p += 5;
	  n++;
	}

	n = (n + 1) >> 1;
	char buf[6];
	sprintf(buf, "%x%x%x%x ", n >> 12, 0xF & (n >> 8), 0xF & (n >> 4), n & 0xF);
    strncpy(rv + 15, buf, 4);
	rv[19] = ' ';
	return rv;
}

char *pronto2uirt(char *pronto)
{
    int len[2];
	int i, w;
    sscanf(pronto + 10, "%x", &len[0]);
	sscanf(pronto + 15, "%x", &len[1]);
	char *p = pronto + 20;
	static char rv[1000];
	char *q = rv;
	for (i = 0; i < 2; i++)
	{	  
	  if (len[i])
	  {
	    *q = 'R';
	    q++;
	  }
	  while (len[i])
	  {
		sscanf(p, "%x", &w);
		if (w > 0x7f)
		{
			sprintf(q, "%2x", 0x80 | (w >> 8));
			q += 2;
		}
		sprintf(q, "%2x", w & 0x7f);
		p += 5;
		len[i]--;
	  }
	  if (i == 0)
	  {
	    *q = '\n';
	    q++;
	  }
	}
    *q = '\0';

	return rv;
}

int sapien_guess_code(char *uirt)
{
	int code = 0;
	//process the uirt data
	if (uirt[0] != 'R') return -1;

	char *q = uirt + 1;
	int val, i;
    for (i = 0; i < 8; i++)
	  if (! *(q++)) return -1;

    while ((*q != 0) && (*q != '\n'))
    {
	  int val2, j;
	  for (j = 0; j < 2; j++)
	  {
	    val = 0;
	    if ((*q == 0) || (*(q + 1) == 0)) break;

	    val = (hexdigit(q) << 4) | hexdigit(q + 1);
	    q += 2;
	    if (val > 0x7f) 
	    {
		  if ((*q == 0) || (*(q + 1) == 0)) break;
		  val = ((val & 0x7f) << 8) | 
		        (hexdigit(q) << 4) | hexdigit(q + 1);
		  q += 2;
	    }
		if (!j) val2 = val;
	  }
	  val = (val2 << 8) | val;

	  int bit;
	  if (val < 0x3000) bit = 0;
	  else bit = 1;

	  code <<= 1;
	  code |= bit;
	}

	return code;
}

int main(int argc, char* argv[])
{
  int i;

  if (argc == 2)
    if ((strcmp(argv[1], "-?") == 0) ||
        (strcmp(argv[1], "-h") == 0) ||
		(strcmp(argv[1], "--help") == 0) ||
		(strcmp(argv[1], "/?") == 0)) argc = 1;

  if (argc == 2)
  {
    if (strcmp(argv[1], "-commands") == 0)
	{
	  printf("\nThe following commands are recognized for Sapien v1:\n");
	  for (i = 0; i < 128; i++)
	    if (cmd[i][0] != '-') printf("           %s (%x)\n", cmd[i], 128 + i);
	  return 0;
    }
	else if (strcmp(argv[1], "-commandsv2") == 0)
	{
	  printf("\nThe following commands are recognized for Sapien v2:\n");
	  for (i = 0; i < SAPIEN2_CMD_COUNT; i++)
	    if (cmd2[i][0] != '_') printf("           %s (%x)\n", cmd2[i], codes2[i]);
		else printf("           DX%s (%x)\n", cmd2[i], codes2[i]);
      printf("\nIn the commands starting with DX, replace X with direction: \n");
      printf("  0 - Up/Forward\n  1 - Down/Backward\n  2 - Left\n  3 - Right\n  4 - Up and Left\n  5 - Up and Right\n  6 - Down and Left\n");
	  printf("  7 - Down and Right\n");
	  return 0;
    }
	else argc = 1;
  }

  if (argc == 1)
  {
	  printf("\nusage: ir2ir from_type to_type command_or_code\n\n");
	  printf("        from_type and to_type can be one of the following:\n");
	  printf("         -v1hex    sapien v1 hex code (8 bit)\n");
	  printf("         -v2hex    sapien v2 hex code (12 bit)\n");
	  printf("         -v1cmd    sapien v1 human-readable command string\n");
	  printf("         -v2cmd    sapien v2 human-readable command string\n");
	  printf("         -uirt     UIRT device specific format (UUIRTDRV_IRFMT_UUIRT)\n");
	  printf("         -pronto   universal Pronto IR code format (UUIRTDRV_IRFMT_PRONTO)\n");
	  printf("        to view the list of commands for Sapien v1, run ir2ir -commands\n");
	  printf("        to view the list of commands for Sapien v2, run ir2ir -commandsv2\n");
	  return 0;
  }

  int t[3];
  int code;

  for (i = 1; i <= 2; i++)
  {
    if (strcmp(argv[i], "-v1hex") == 0) t[i] = TYPE_V1HEX;
    else if (strcmp(argv[i], "-v2hex") == 0) t[i] = TYPE_V2HEX;
    else if (strcmp(argv[i], "-v1cmd") == 0) t[i] = TYPE_V1CMD;
    else if (strcmp(argv[i], "-v2cmd") == 0) t[i] = TYPE_V2CMD;
    else if (strcmp(argv[i], "-uirt") == 0) t[i] = TYPE_UIRT;
	else if (strcmp(argv[i], "-pronto") == 0) t[i] = TYPE_PRONTO;
	else 
	{
		printf("ir2ir: unrecognized type %s\n", argv[i]);
		return 1;
	}
  }

  if ((((t[1] == TYPE_V1CMD) || (t[1] == TYPE_V1HEX)) &&
       ((t[2] == TYPE_V2CMD) || (t[2] == TYPE_V2HEX))) || 
      (((t[1] == TYPE_V2CMD) || (t[1] == TYPE_V2HEX)) &&
       ((t[2] == TYPE_V1CMD) || (t[2] == TYPE_V1HEX))))
  {
	  printf("ir2ir: unsupported combination\n");
	  return 1;
  }
  double freq;
  if (t[1] == TYPE_PRONTO)
  {
    sscanf(argv[3] + 5, "%x", &freq);
	freq = 4145146.0 / (double)freq;
  }

  if (((t[1] == TYPE_V1HEX) || (t[1] == TYPE_V2HEX)) && ((argv[3][0] != '0') || (argv[3][1] != 'x')))
  {
	  printf("ir2ir: argument %s: expected hex-constant starting with 0x\n", argv[3]);
      return 1;
  }

  if ((t[1] == TYPE_V1CMD) && (t[2] == TYPE_V1HEX))
	printf("0x%x\n", sapien_decode(argv[3], 1));
  else if ((t[1] == TYPE_V2CMD) && (t[2] == TYPE_V2HEX))
    printf("0x%x\n", sapien_decode(argv[3], 2));
  else if ((t[1] == TYPE_V1CMD) && (t[2] == TYPE_UIRT))
    printf("%s\n", sapienhex2uirt(sapien_decode(argv[3], 1), 8));
  else if ((t[1] == TYPE_V2CMD) && (t[2] == TYPE_UIRT))
    printf("%s\n", sapienhex2uirt(sapien_decode(argv[3], 2), 12));
  else if ((t[1] == TYPE_V1HEX) && (t[2] == TYPE_UIRT))
  {
    sscanf(argv[3] + 2, "%x", &code);
    printf("%s\n", sapienhex2uirt(code, 8));
  }
  else if ((t[1] == TYPE_V2HEX) && (t[2] == TYPE_UIRT))
  {
    sscanf(argv[3] + 2, "%x", &code);
	printf("%s\n", sapienhex2uirt(code, 12));
  }
  else if ((t[1] == TYPE_V1HEX) && (t[2] == TYPE_V1CMD))
  {
    sscanf(argv[3] + 2, "%x", &code);
	if ((code < 128) || (code > 255))
	  printf("unrecongized code\n");
	else
      printf("%s\n", cmd[code - 128]);
  }
  else if ((t[1] == TYPE_UIRT) && (t[2] == TYPE_V1HEX))
    printf("0x%x\n", sapien_guess_code(argv[3]));
  else if ((t[1] == TYPE_UIRT) && (t[2] == TYPE_V2HEX))
    printf("0x%x\n", sapien_guess_code(argv[3]));
  else if ((t[1] == TYPE_UIRT) && (t[2] == TYPE_V1CMD))
  {
    code = sapien_guess_code(argv[3]) - 128;
	if ((code >= 0) && (code < 128)) printf("%s", cmd[code]);
	else printf("not recognized\n");
  }
  else if (((t[1] == TYPE_UIRT) || (t[1] == TYPE_V2HEX) || (t[1] == TYPE_PRONTO)) && (t[2] == TYPE_V2CMD))
  {
    if (t[1] == TYPE_UIRT) code = sapien_guess_code(argv[3]);
	else if (t[1] == TYPE_V2HEX) sscanf(argv[3] + 2, "%x", &code);
	else 
	{
      if (fabs(freq - 39200.0) > 1500) printf("ir2ir: warning, unexpected carrier frequency %lf\n", freq);
      code = sapien_guess_code(pronto2uirt(argv[3]));
	}
	for (i = 0; i < SAPIEN2_CMD_COUNT; i++)
	    if (code == codes2[i]) 
		{
			printf("%s\n", cmd2[i]);
			break;
		}
		else if ((cmd2[i][0] == '_') && (codes2[i] == (code & (~7))))
		{
			printf("D%i%s\n", code & 7, cmd2[i]);
			break;
		}
	if (i == SAPIEN2_CMD_COUNT) printf("not recognized\n");	  
  }
  else if ((t[1] == TYPE_V1HEX) && (t[2] == TYPE_PRONTO))
  {
    sscanf(argv[3] + 2, "%x", &code);
    printf("%s\n", hex2pronto(code, 1));
  }
  else if ((t[1] == TYPE_V2HEX) && (t[2] == TYPE_PRONTO))
  {
    sscanf(argv[3] + 2, "%x", &code);
	printf("%s\n", hex2pronto(code, 2));
  }
  else if ((t[1] == TYPE_V1CMD) && (t[2] == TYPE_PRONTO))
    printf("%s\n", hex2pronto(sapien_decode(argv[3], 1), 1));
  else if ((t[1] == TYPE_V2CMD) && (t[2] == TYPE_PRONTO))
    printf("%s\n", hex2pronto(sapien_decode(argv[3], 2), 2));
  else if ((t[1] == TYPE_UIRT) && (t[2] == TYPE_PRONTO))
  {
	  printf("ir2ir: assuming 39.2 kHz\n");
	  printf("%s\n", uirt2pronto(argv[3]));
  }
  else if ((t[1] == TYPE_PRONTO) && (t[2] == TYPE_UIRT))
  {
	printf("ir2ir: the frequency is %lfHz\n", freq); 
    printf("%s\n", pronto2uirt(argv[3]));
  }
  else if ((t[1] == TYPE_PRONTO) && (t[2] == TYPE_V1HEX))
  {
	if (fabs(freq - 39200.0) > 1500) printf("ir2ir: warning, unexpected carrier frequency %lf\n", freq);
	printf("0x%x\n", sapien_guess_code(pronto2uirt(argv[3])));
  }
  else if ((t[1] == TYPE_PRONTO) && (t[2] == TYPE_V1CMD))
  {
	if (fabs(freq - 39200.0) > 1500) printf("ir2ir: warning, unexpected carrier frequency %lf\n", freq);
    code = sapien_guess_code(pronto2uirt(argv[3])) - 128;
	if ((code >= 0) && (code < 128))
	  printf("%s\n", cmd[code]);
	else printf("not recognized\n");
  }
  else if ((t[1] == TYPE_PRONTO) && (t[2] == TYPE_V2HEX))
  {
	if (fabs(freq - 39200.0) > 1500) printf("ir2ir: warning, unexpected carrier frequency %lf\n", freq);
	printf("0x%x\n", sapien_guess_code(pronto2uirt(argv[3])));
  }
  else
    printf("unsupported combination\n");
  return 0;
}

